@@ -25,13 +25,13 @@ class TorchNet(object):
         self._devices = ['cuda:%d' % k if k is not None and torch.cuda.is_available() and k != 'cpu' else 'cpu' for k in idlist]
         #self._tensortype = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor       
         self._tensortype = torch.FloatTensor       
-        
-        if not hasattr(self, '_gpulist') or not hasattr(self, '_models') or idlist != self._gpulist  or not hasattr(self, '_models'):
+
+        if not hasattr(self, '_gpulist') or not hasattr(self, '_models') or idlist != self._gpulist  or not hasattr(self, '_models'):        
             self._models = [copy.deepcopy(self._model).to(d, non_blocking=False) for d in self._devices]
             for (d,m) in zip(self._devices, self._models):
                 m.eval()
             self._gpulist = idlist
-        torch.set_grad_enabled(False)            
+        torch.set_grad_enabled(False)
         return self
 
     def cpu(self, batchsize=None):
@@ -111,7 +111,7 @@ class Yolov5(TorchNet):
         self._index2cls = {k:c for (c,k) in self._cls2index.items()}
 
         self._device = None
-        self._gpulist = gpu
+        #self._gpulist = gpu  # will be set in self.gpu()
         if gpu is not None:
             self.gpu(gpu, batchsize)
         else:
@@ -128,19 +128,20 @@ class Yolov5(TorchNet):
         assert objects is None or (isinstance(objects, list) and all([(k[0] if isinstance(k, tuple) else k) in self._cls2index for k in objects])), "Objects must be a list of allowable categories"
         objects = {(k[0] if isinstance(k,tuple) else k):(k[1] if isinstance(k,tuple) else k) for k in objects} if isinstance(objects, list) else objects
 
-        imlist = tolist(imlist)
-        imlistdets = []
-        t = torch.cat([im.clone(shallow=True).maxsquare().mindim(self._mindim).gain(1.0/255.0).torch() for im in imlist])  # triggers load
-        if torch.cuda.is_available() and not self.iscpu():
-            t = t.pin_memory()
+        with torch.no_grad():
+            imlist = tolist(imlist)
+            imlistdets = []
+            t = torch.cat([im.clone(shallow=True).maxsquare().mindim(self._mindim).gain(1.0/255.0).torch(order='NCHW') for im in imlist])  # triggers load
+            if torch.cuda.is_available() and not self.iscpu():
+                t = t.pin_memory()
 
-        assert len(t) <= self.batchsize(), "Invalid batch size"
-        todevice = [b.to(d, memory_format=torch.contiguous_format, non_blocking=True) for (b,d) in zip(t.split(self._batchsize), self._devices)]  # contiguous_format required for torch-1.8.1
-        fromdevice = [m(b)[0] for (m,b) in zip(self._models, todevice)]     # detection
+            assert len(t) <= self.batchsize(), "Invalid batch size: %d > %d" % (len(t), self.batchsize())
+            todevice = [b.to(d, memory_format=torch.contiguous_format, non_blocking=True) for (b,d) in zip(t.split(self._batchsize), self._devices)]  # contiguous_format required for torch-1.8.1
+            fromdevice = [m(b)[0] for (m,b) in zip(self._models, todevice)]     # detection
         
-        t_out = [torch.squeeze(t, dim=0) for d in fromdevice for t in torch.split(d, 1, 0)]   # unpack batch to list of detections per imag
-        t_out = [torch.cat((t[:,0:5], torch.argmax(t[:,5:], dim=1, keepdim=True)), dim=1) for t in t_out]  # filter argmax on device 
-        t_out = [t[t[:,4]>conf].cpu().numpy() for t in t_out]  # filter conf on device (this must be last)
+            t_out = [torch.squeeze(t, dim=0) for d in fromdevice for t in torch.split(d, 1, 0)]   # unpack batch to list of detections per imag
+            t_out = [torch.cat((t[:,0:5], torch.argmax(t[:,5:], dim=1, keepdim=True)), dim=1) for t in t_out]  # filter argmax on device 
+            t_out = [t[t[:,4]>conf].cpu().detach().numpy() for t in t_out]  # filter conf on device (this must be last)
 
         k_valid_objects = set([self._cls2index[k] for k in objects.keys()]) if objects is not None else self._cls2index.values()        
         for (im, dets) in zip(imlist, t_out):
@@ -209,8 +210,8 @@ class Yolov3(TorchNet):
 
         imlist = tolist(im)
         imlistdets = []
-        t = torch.cat([im.clone().maxsquare().mindim(self._mindim).gain(1.0/255.0).torch() for im in imlist]).type(self._tensortype)  # triggers load
-        t_out = super().__call__(t).numpy()   # parallel multi-GPU evaluation, using TorchNet()
+        t = torch.cat([im.clone().maxsquare().mindim(self._mindim).gain(1.0/255.0).torch(order='NCHW') for im in imlist]).type(self._tensortype)  # triggers load
+        t_out = super().__call__(t).detach().numpy()   # parallel multi-GPU evaluation, using TorchNet()
         for (im, dets) in zip(imlist, t_out):
             k_class = np.argmax(dets[:,5:], axis=1).flatten().tolist()
             k_det = np.argwhere((dets[:,4] > conf).flatten() & np.array([((objects is None) or (self._index2cls[k] in objects.keys())) for k in k_class])).flatten().tolist()
@@ -350,7 +351,8 @@ class MultiscaleVideoTracker(MultiscaleObjectDetector):
         self._gate = gate
 
     def _track(self, vi, stride=1):
-        """Yield vipy.video.Scene(), an incremental tracked result for each frame"""
+        """Yield vipy.video.Scene(), an incremental tracked result for each frame.
+        """
         assert isinstance(vi, vipy.video.Video), "Invalid input"
 
         (det, n) = (super().__call__, self._mindim)
@@ -387,7 +389,7 @@ class VideoProposal(Proposal):
        Track-based object proposals in video.
     """
     def allowable_objects(self):
-        return ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike']
+        return ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike', 'truck']
 
     def isallowable(self, v):
         assert isinstance(v, vipy.video.Video), "Invalid input - must be vipy.video.Video not '%s'" % (str(type(v)))
@@ -421,10 +423,11 @@ class VideoProposal(Proposal):
         tensor = vc.flush().crop(bb, zeropad=False).maxsquare().mindim(self._mindim).torch()[::dt]  # transformed video, NxCxHxW, re-triggers load due to crop()
 
         for i in range(0, len(tensor), self.batchsize()):
-            t = tensor[i:i+self.batchsize()]
-            todevice = [b.to(d, non_blocking=True) for (b,d) in zip(t.split(self._batchsize) , self._devices)]  # async?
-            fromdevice = [m(b)[0] for (m,b) in zip(self._models, todevice)]     # detection!
-            dets = [torch.squeeze(t, dim=0).cpu().numpy() for d in fromdevice for t in torch.split(d, 1, 0)]   # unpack batch to list of detections per imag
+            with torch.no_grad():
+                t = tensor[i:i+self.batchsize()]
+                todevice = [b.to(d, non_blocking=True) for (b,d) in zip(t.split(self._batchsize) , self._devices)]  # async?
+                fromdevice = [m(b)[0] for (m,b) in zip(self._models, todevice)]     # detection!
+                dets = [torch.squeeze(t, dim=0).cpu().detach().numpy() for d in fromdevice for t in torch.split(d, 1, 0)]   # unpack batch to list of detections per imag
 
             for (j, det) in enumerate(dets):
                 # Objects in transformed video
@@ -520,16 +523,17 @@ class ActorAssociation(MultiscaleVideoTracker):
 
     @staticmethod
     def isallowable(v, actor_class, association_class, fps=None):
-        allowable_objects = ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike']        
+        allowable_objects = ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike', 'truck']        
         return (actor_class.lower() in allowable_objects and
-                association_class.lower() in allowable_objects and
+                all([a.lower() in allowable_objects for a in vipy.util.tolist(association_class)]) and
                 actor_class.lower() in v.objectlabels(lower=True))
         
 
-    def __call__(self, v, actor_class, association_class, fps=None, dilate=2.0, activity_class=None, maxcover=0.8, max_associations=1):
-        allowable_objects = ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike']        
+    def __call__(self, v, actor_class, association_class, fps=None, dilate=2.0, activity_class=None, maxcover=0.8, max_associations=1, min_confidence=0.4):
+        allowable_objects = ['person', 'vehicle', 'car', 'motorcycle', 'object', 'bicycle', 'motorbike', 'truck']        
+        association_class = [a.lower() for a in vipy.util.tolist(association_class)]
         assert actor_class.lower() in allowable_objects, "Primary Actor '%s' not in allowable target class '%s'" % (actor_class.lower(), str(allowable_objects))
-        assert association_class.lower() in allowable_objects, "Actor Association '%s' not in allowable target class '%s'" % (association_class.lower(), str(allowable_objects))
+        assert all([a in allowable_objects for a in allowable_objects]), "Actor Association '%s' not in allowable target class '%s'" % (str(association_class), str(allowable_objects))
         assert actor_class.lower() in v.objectlabels(lower=True), "Actor Association can only be performed with scenes containing an allowable actor not '%s'" % str(v.objectlabels())
         
         # Track objects
@@ -539,15 +543,15 @@ class ActorAssociation(MultiscaleVideoTracker):
                 t._framerate = v.framerate()  # HACK: backwards compatibility
             for a in vc.activities().values():
                 a._framerate = v.framerate()  # HACK: backwards compatibility
-
-        vt = self.track(vc.framerate(fps).clone()).framerate(v.framerate()) if fps is not None else self.track(vc.clone())
+        vc = vc.framerate(fps) if fps is not None else vc   # downsample
+        vt = self.track(vc.clone())  # track at downsampled framerate
 
         # Actor assignment: for every activity, find track with best target object assignment to actor (first track in video)
         for a in vc.activities().values():
-            candidates = [t for t in vt.tracks().values() if (t.category().lower() == association_class.lower() and
+            candidates = [t for t in vt.tracks().values() if (t.category().lower() in association_class and
                                                               t.during_interval(a.startframe(), a.endframe()) and
-                                                              t.confidence() > 0.4 and  # must have minimum confidence
-                                                              (association_class.lower() != actor_class.lower() or t.segmentcover(vc.actor()) < maxcover) and
+                                                              t.confidence() > min_confidence and  # must have minimum confidence
+                                                              (actor_class.lower() not in association_class or t.segmentcover(vc.actor()) < maxcover) and
                                                               vc.actor().boundingbox().dilate(dilate).hasintersection(t.boundingbox()))] # candidate assignment (cannot be actor, or too far from actor)
             if len(candidates) > 0:
                 # best assignment is track closest to actor with maximum confidence and minimum dilated overlap
@@ -556,7 +560,8 @@ class ActorAssociation(MultiscaleVideoTracker):
                     if a.during_interval(t.startframe(), t.endframe()) and activity_class is None or a.category() == activity_class:
                         a.add(t)
                         vc.add(t)
-        return vc
+
+        return vc.framerate(v.framerate()) if vc.framerate() != v.framerate() else vc   # upsample
 
     
 def _collectorproposal_vs_objectproposal(v, dt=1, miniou=0.2, smoothing='spline'):
